//f.olofsson2018
//Generic ESP8266 Module, 80MHz
//use together with tamas3_teensy

//TODO udp broadcast bug - is sluggish
//TODO blink debug led?

//--osc protocol:
//send to laptop
// \ready id
// \dig id val val val val val val val val val val val val val val val val      //dig (0/1)
// \ana id val val val val val val val val val val val val val val val val      //ana (0-255)
//receive from laptop
// \led num val           //led num (0-15), val (0-255)
// \leds val val val val val val val val val val val val val val val val        //leds (0-255)

//--serial protocol:
//send to teensy
// 150 151 num val 152    //led (0-15, 0-255)
// 175 176 val val val val val val val val val val val val val val val val 177  //leds (0-255)
//receive from teensy
// 0 1 id 2               //id
// 200 201 val val val val val val val val val val val val val val val val 202  //ana (0-255)
// 250 251 lo hi 252      //dig (as bits)

#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <OSCMessage.h>
#include <OSCData.h>

#define SOFTAP 1  //EDIT client or soft access point
#define PORT 14999  //EDIT input osc port
unsigned int outPort = 57120;  //EDIT output osc port (supercollider by default)
const char *ssid = "fIGMA01"; //EDIT softAccessPoint network name
const char *password = "fIGMA12345";  //EDIT password
const char *espname = "fIGMA";
#define STATICIP 0  //EDIT set to 1 or 0
#if STATICIP==1
IPAddress ip(192, 168, 1, 50); //EDIT optional static ip
IPAddress gateway(192, 168, 1, 1);  //EDIT optional router
IPAddress subnet(255, 255, 255, 0);
#endif

#define NUMLED 16
#define NUMDIG 16
#define NUMANA 16
IPAddress outIp;
WiFiUDP Udp;
int id;  //received from teensy
byte state = 0; //for decoding serial input
int anaLast[NUMANA];

void setup() {
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, HIGH); //debug
  delay(10);
  Serial.begin(1000000);
#if SOFTAP==1
  WiFi.mode(WIFI_AP);
#if STATICIP==1
  WiFi.softAPConfig(ip, gateway, subnet);
#endif
  WiFi.softAP(ssid, password);
#else
  WiFi.mode(WIFI_STA);
  WiFi.hostname(espname);
  WiFi.begin(ssid, password);
#if STATICIP==1
  WiFi.config(ip, gateway, subnet);
#endif
  while (WiFi.status() != WL_CONNECTED) {
    delay(100); //TODO change this so that it can timeout
  }
#endif
#if SOFTAP==1
  outIp = WiFi.softAPIP();
#else
  outIp = WiFi.localIP();
#endif
  outIp[3] = 255;  //by default use broadcast ip x.x.x.255
  Udp.begin(PORT);
  id = serialReceiveID();
  oscReady();
  digitalWrite(LED_BUILTIN, LOW); //debug
}

int serialReceiveID() {
  int i;
  while (Serial.available() < 4) {}
  if (Serial.read() == 0 && Serial.read() == 1) {
    i = Serial.read();
    if (Serial.read() != 2) {
      //while (1) {}
      i = -1;  //indicate error - did not get id from teensy
    }
  }
  return i;
}
void oscReady() {
  OSCMessage msg("/ready");
  msg.add(id);
  Udp.beginPacket(outIp, outPort);
  msg.send(Udp);
  Udp.endPacket();
  yield();
  msg.empty();
}
void oscLed(OSCMessage &msg) {
  if (msg.size() == 2) {
    Serial.write(150);
    Serial.write(151);
    int num = getIntData(msg, 0);
    int val = getIntData(msg, 1);
    num = constrain(num, 0, NUMLED - 1);
    val = constrain(val, 0, 255);
    Serial.write(num);
    Serial.write(val);
    Serial.write(152);
    Serial.flush();
  }
}
void oscLeds(OSCMessage &msg) {
  if (msg.size() == NUMLED) {
    Serial.write(175);
    Serial.write(176);
    for (byte num = 0; num < NUMLED; num++) {
      int val = getIntData(msg, num);
      val = constrain(val, 0, 255);
      Serial.write(val);
    }
    Serial.write(177);
    Serial.flush();
  }
}
void oscIp(OSCMessage &msg) {
  outIp[0] = getIntData(msg, 0);
  outIp[1] = getIntData(msg, 1);
  outIp[2] = getIntData(msg, 2);
  outIp[3] = getIntData(msg, 3);
}
void oscPort(OSCMessage &msg) {
  outPort = getIntData(msg, 0);
}
void oscOutputDig(byte lo, byte hi) {
  OSCMessage msg("/dig");
  msg.add(id);
  for (byte i = 0; i < 8; i++) {
    msg.add((lo >> i) & 1);
  }
  for (byte i = 0; i < 8; i++) {
    msg.add((hi >> i) & 1);
  }
  Udp.beginPacket(outIp, outPort);
  msg.send(Udp);
  Udp.endPacket();
  yield();
  msg.empty();
}
void oscOutputAna() {
  OSCMessage msg("/ana");
  msg.add(id);
  for (byte num = 0; num < NUMANA; num++) {
    msg.add(anaLast[num]);
  }
  Udp.beginPacket(outIp, outPort);
  //Udp.beginPacketMulticast(outIp, outPort, WiFi.localIP());
  msg.send(Udp);
  Udp.endPacket();
  yield();
  msg.empty();
}
int getIntData(OSCMessage &msg, int val) {
  if (msg.isInt(val)) {
    return msg.getInt(val);
  }
  return int(msg.getFloat(val));
}

void loop() {
  OSCMessage oscMsg;
  int packetSize = Udp.parsePacket();
  if (packetSize) {
    while (packetSize--) {
      oscMsg.fill(Udp.read());
    }
    if (!oscMsg.hasError()) {
      oscMsg.dispatch("/led", oscLed);
      oscMsg.dispatch("/leds", oscLeds);
      oscMsg.dispatch("/ip", oscIp);
      oscMsg.dispatch("/port", oscPort);
    }
  }
  while (Serial.available() > 0) {
    byte data;
    switch (state) {
      case 0:
        data = Serial.read();
        if (data == 200) {
          state = 200;
        } else if (data == 250) {
          state = 250;
        }
        break;
      case 200:
        if (Serial.read() == 201) {
          state = 201;
        } else {
          state = 0;
        }
        break;
      case 201:
        if (Serial.available() > NUMANA) {
          for (byte num = 0; num < NUMANA; num++) {
            anaLast[num] = Serial.read();
          }
          if (Serial.read() == 202) {
            oscOutputAna();
          }
          state = 0;
        }
        break;
      case 250:
        if (Serial.read() == 251) {
          state = 251;
        } else {
          state = 0;
        }
        break;
      case 251:
        if (Serial.available() >= 3) {
          byte lo = Serial.read();
          byte hi = Serial.read();
          if (Serial.read() == 252) {
            oscOutputDig(lo, hi);
          }
          state = 0;
        }
        break;
    }
  }
}
