//f.olofsson2016-2018
//for Teensy 3.5 (120MHz, Serial + MIDI)
//and a tlc5940 and an optional esp8266

//TODO:
//gamma correction
//Serial1 id check

//xlat 3 (dip24) - teensy 3 (dip5)
//blank 2 (dip23) - teensy 4 (dip6)
//gsclk 25 (dip18) - teensy 5 (dip7)
//sin 5 (dip26) - teensy 6 (dip8)
//sclk 4 (dip25) - teensy 7 (dip9)
//iref 27 (dip20) - 2k2 to gnd
//dcprg (dip19) - 5v
//vcc (dip21) - 5v
//gnd (dip22) - gnd
//vprg (dip27) - gnd
//xerr (dip16) - x

//--serial protocol:
//send to laptop
// 200 201 val val val val val val val val val val val val val val val val 202  //ana (0-255)
// 250 251 lo hi 252        //dig (as bits)
//receive from laptop
// 150 151 num val 152      //led (0-15, 0-255)
// 175 176 val val val val val val val val val val val val val val val val 177  //leds (0-255)

//--serial1 protocol:
//send to esp
// 0 1 id 2                 //id
// 200 201 val val val val val val val val val val val val val val val val 202  //ana (0-255)
// 250 251 lo hi 252        //dig (as bits)
//receive from esp
// 150 151 num val 152      //led (0-15, 0-255)
// 175 176 val val val val val val val val val val val val val val val val 177  //leds (0-255)

//--midi protocol:
//send to laptop
// num val chan      //ana (controlChange 20-35)
// note vel chan     //dig (noteOn noteOff 60-75, vel 99/0)
//receive from laptop
// num val chan      //led (controlChange 50-65, vel 0-127)

#include "Tlc5940.h"
#include <Bounce.h>

#define ID 1  //NOTE: also change in name.c - will be midi channel, midi device name, osc prefix

#define DEBOUNCETIME 10   //time in milliseconds for digital inputs
#define SAMPLETIME 20     //minimum time in milliseconds between analog inputs
#define MIDINOTEOFFSET 60 //for digital inputs
#define MIDICTRLOFFSET 20 //for analog inputs
#define MIDICTRLOUTOFFSET 40 //for led outputs
#define NUMLED 16
#define NUMDIG 16
#define NUMANA 16

const byte digPins[] = {2, 8, 9, 10, 11, 12, 33, 24, 25, 26, 27, 28, 29, 30, 31, 32};
const byte anaPins[] = {18, 17, 16, 15, 14, 19, 20, 21, 22, 39, 38, 37, 36, 35, 34, 23};
const byte ledPins[] = {7, 8, 9, 10, 11, 12, 13, 14, 0, 1, 2, 3, 4, 5, 6, 15};
const byte ledPin = 13; //debug

Bounce *bounces[NUMDIG];
byte digLast[NUMDIG];
byte anaLast[NUMANA];
byte midLast[NUMANA];
unsigned long nextTime;
byte states[] = {0, 0}; //for decoding serial inputs (usb, esp)

void setup() {
  Serial.begin(115200); //usb
  Serial1.begin(1000000); //esp
  for (byte i = 0; i < NUMDIG; i++) {
    bounces[i] = new Bounce(digPins[i], DEBOUNCETIME);
    pinMode(digPins[i], INPUT_PULLUP);
    digLast[i] = 0;
  }
  Tlc.init();
  tlcStartup();
  nextTime = 0;
  pinMode(ledPin, OUTPUT);  //used for debugging
  digitalWrite(ledPin, HIGH);
  Serial1.write(0);
  Serial1.write(1);
  Serial1.write(ID);  //TODO check if this is reliable - wait?
  Serial1.write(2);
  Serial1.flush();
  usbMIDI.setHandleControlChange(midiInput);
}

void midiInput(byte channel, byte ctrl, byte value) {
  if (ctrl >= MIDICTRLOUTOFFSET && ctrl <= (MIDICTRLOUTOFFSET + 15)) {
    ledSet(ctrl - MIDICTRLOUTOFFSET, value << 5); //scale up from 0-127 to 0-4095
    ledUpdate();
  }
}
void serialInput(Stream &port, byte stateIndex) {
  while (port.available() > 0) {
    byte data;
    switch (states[stateIndex]) {
      case 0:
        data = port.read();
        if (data == 150) {
          states[stateIndex] = 150;
        } else if (data == 175) {
          states[stateIndex] = 175;
        }
        break;
      case 150:
        if (port.read() == 151) {
          states[stateIndex] = 151;
        } else {
          states[stateIndex] = 0;
        }
        break;
      case 151:
        if (port.available() >= 3) {
          byte num = port.read();
          data = port.read();
          ledSet(num, data << 4); //scale up from 0-255 to 0-4095
          if (port.read() == 152) {
            ledUpdate();
          }
          states[stateIndex] = 0;
        }
        break;
      case 175:
        if (port.read() == 176) {
          states[stateIndex] = 176;
        } else {
          states[stateIndex] = 0;
        }
        break;
      case 176:
        if (port.available() > NUMLED) {
          for (byte num = 0; num < NUMLED; num++) {
            data = port.read();
            ledSet(num, data << 4); //scale up from 0-255 to 0-4095
          }
          if (port.read() == 177) {
            ledUpdate();
          }
          states[stateIndex] = 0;
        }
        break;
    }
  }
}
void ledSet(byte i, int data) {
  Tlc.set(ledPins[i], data); //TODO gamma correction
}
void ledUpdate() {
  digitalWrite(ledPin, HIGH); //debug
  Tlc.update();
}
void serialOutputDig(Stream &port) {  //send digital inputs as bits
  port.write(250);
  port.write(251);
  for (byte i = 0; i < (NUMDIG / 8); i++) {
    byte val = 0;
    for (byte j = 0; j < 8; j++) {
      val += (1 << j) * digLast[i * 8 + j];
    }
    port.write(val);
  }
  port.write(252);
  port.flush();
}
void serialOutputAna(Stream &port) {  //send analog inputs as bytes
  port.write(200);
  port.write(201);
  for (byte num = 0; num < NUMANA; num++) {
    port.write(anaLast[num]);
  }
  port.write(202);
  port.flush();
}

void loop() {
  bool changed = false;

  //--dig
  for (byte num = 0; num < NUMDIG; num++) {
    bounces[num]->update();
    if (bounces[num]->fallingEdge()) {
      usbMIDI.sendNoteOn(MIDINOTEOFFSET + num, 99, ID);
      digLast[num] = 1;
      changed = true;
    } else if (bounces[num]->risingEdge()) {
      usbMIDI.sendNoteOff(MIDINOTEOFFSET + num, 0, ID);
      digLast[num] = 0;
      changed = true;
    }
  }
  if (changed) {  //only send if something changed
    serialOutputDig(Serial);  //usb
    serialOutputDig(Serial1); //esp
  }

  //--ana
  if (millis() >= nextTime) {
    digitalWrite(ledPin, LOW);  //debug
    changed = false;
    for (byte num = 0; num < NUMANA; num++) {
      byte val = analogRead(anaPins[num]) >> 2; //scale down to 0-255
      if (anaLast[num] != val) {
        byte mid = val >> 1;  //for midi scale down again to 0-127
        if (midLast[num] != mid) {
          usbMIDI.sendControlChange(MIDICTRLOFFSET + num, mid, ID);
          midLast[num] = mid;
        }
        anaLast[num] = val;
        changed = true;
      }
    }
    if (changed) {  //only send if something changed
      serialOutputAna(Serial);  //send to usb
      serialOutputAna(Serial1); //send to esp
      nextTime = millis() + SAMPLETIME; //make sure not to read+send too often
    }
  }

  //--leds
  usbMIDI.read(); //keep midi input alive
  serialInput(Serial, 0);  //check incoming from usb
  serialInput(Serial1, 1); //check incomming from esp
}

void tlcStartup() {  //half wave sine
  for (int j = 0; j < 50; j++) {
    Tlc.clear();
    for (int i = 0; i < NUMLED; i++) {
      Tlc.set(ledPins[i], max(0, sin(i / float(NUMLED) * PI + PI + (j / 49.0 * 2 * PI))) * 4095);
    }
    Tlc.update();
    delay(20);
  }
  Tlc.clear();
  Tlc.update();
}
